@@ -3,7 +3,13 @@ from collections import namedtuple
 
 import aiohttp
 import voluptuous as vol
-from voluptuous import Invalid
+from voluptuous import Invalid, MultipleInvalid
+from voluptuous.humanize import humanize_error
+from solax.utils import (
+    div10, div100, feedin_energy, total_energy, charge_energy, pv_energy,
+    discharge_energy, consumption, twoway_div10, twoway_div100, to_signed,
+    eps_total_energy
+)
 
 
 class InverterError(Exception):
@@ -57,6 +63,13 @@ class Inverter:
         """
         raise NotImplementedError()
 
+    @classmethod
+    def postprocess_map(cls):
+        """
+        Return map of functions to be applied to each sensor value
+        """
+        return {}
+
     @classmethod
     def schema(cls):
         """
@@ -64,13 +77,15 @@ class Inverter:
         """
         raise NotImplementedError()
 
-    @staticmethod
-    def map_response(resp_data, sensor_map):
-        return {
-            sensor_name: resp_data[i]
-            for sensor_name, (i, _)
-            in sensor_map.items()
-        }
+    @classmethod
+    def map_response(cls, resp_data):
+        result = {}
+        for sensor_name, (idx, _) in cls.sensor_map().items():
+            val = resp_data[idx]
+            result[sensor_name] = val
+        for sensor_name, processor in cls.postprocess_map().items():
+            result[sensor_name] = processor(result[sensor_name], result)
+        return result
 
 
 async def discover(host, port, pwd='') -> Inverter:
@@ -167,7 +182,7 @@ class XHybrid(Inverter):
         json_response = json.loads(formatted)
         response = cls.schema()(json_response)
         return InverterResponse(
-            data=cls.map_response(response['Data'], cls.__sensor_map),
+            data=cls.map_response(response['Data']),
             serial_number=response['SN'],
             version=response['version'],
             type=response['type']
@@ -199,10 +214,20 @@ class InverterPost(Inverter):
                 resp = await req.read()
         raw_json = resp.decode("utf-8")
         json_response = json.loads(raw_json)
-        response = cls.schema()(json_response)
+        response = {}
+        try:
+            response = cls.schema()(json_response)
+        except (Invalid, MultipleInvalid) as ex:
+            _ = humanize_error(json_response, ex)
+            # print(_)
+            raise
+        if 'SN' in response:
+            serial_number = response['SN']
+        else:
+            serial_number = response['sn']
         return InverterResponse(
-            data=cls.map_response(response['Data'], cls.sensor_map()),
-            serial_number=response['SN'],
+            data=cls.map_response(response['Data']),
+            serial_number=serial_number,
             version=response['ver'],
             type=response['type']
         )
@@ -287,6 +312,112 @@ class X3(InverterPost):
         return cls.__schema
 
 
+class X3V34(InverterPost):
+    """X3 v2.034.06"""
+    __schema = vol.Schema({
+        vol.Required('type'): vol.All(int, 5),
+        vol.Required('sn'): str,
+        vol.Required('ver'): str,
+        vol.Required('Data'): vol.Schema(
+            vol.All(
+                [vol.Coerce(float)],
+                vol.Length(min=200, max=200),
+                )
+            ),
+        vol.Required('Information'): vol.Schema(
+            vol.All(
+                vol.Length(min=10, max=10)
+                )
+            ),
+    }, extra=vol.REMOVE_EXTRA)
+
+    __sensor_map = {
+        'Network Voltage Phase 1':               (0,   'V', div10),
+        'Network Voltage Phase 2':               (1,   'V', div10),
+        'Network Voltage Phase 3':               (2,   'V', div10),
+
+        'Output Current Phase 1':                (3,   'A', twoway_div10),
+        'Output Current Phase 2':                (4,   'A', twoway_div10),
+        'Output Current Phase 3':                (5,   'A', twoway_div10),
+
+        'Power Now Phase 1':                     (6,   'W', to_signed),
+        'Power Now Phase 2':                     (7,   'W', to_signed),
+        'Power Now Phase 3':                     (8,   'W', to_signed),
+
+        'PV1 Voltage':                           (9,   'V', div10),
+        'PV2 Voltage':                           (10,  'V', div10),
+        'PV1 Current':                           (11,  'A', div10),
+        'PV2 Current':                           (12,  'A', div10),
+        'PV1 Power':                             (13,  'W'),
+        'PV2 Power':                             (14,  'W'),
+
+        'Total PV Energy':                       (89,  'kWh', pv_energy),
+        'Total PV Energy Resets':                (90,  ''),
+        'Today\'s PV Energy':                    (112,  'kWh', div10),
+
+        'Grid Frequency Phase 1':                (15,  'Hz', div100),
+        'Grid Frequency Phase 2':                (16,  'Hz', div100),
+        'Grid Frequency Phase 3':                (17,  'Hz', div100),
+
+        'Total Energy':                          (19,  'kWh', total_energy),
+        'Total Energy Resets':                   (20,  ''),
+        'Today\'s Energy':                       (21,  'kWh', div10),
+
+        'Battery Voltage':                       (24,  'V', div100),
+        'Battery Current':                       (25,  'A', twoway_div100),
+        'Battery Power':                         (26,  'W', to_signed),
+        'Battery Temperature':                   (27,  'C'),
+        'Battery Remaining Capacity':            (28,  '%'),
+
+        'Total Battery Discharge Energy':        (30,  'kWh',
+                                                  discharge_energy),
+        'Total Battery Discharge Energy Resets': (31,  ''),
+        'Today\'s Battery Discharge Energy':     (113,  'kWh', div10),
+        'Battery Remaining Energy':              (32,  'kWh', div10),
+        'Total Battery Charge Energy':           (87,  'kWh', charge_energy),
+        'Total Battery Charge Energy Resets':    (88,  ''),
+        'Today\'s Battery Charge Energy':        (114,  'kWh', div10),
+
+        'Exported Power':                        (65,  'W', to_signed),
+        'Total Feed-in Energy':                  (67,  'kWh', feedin_energy),
+        'Total Feed-in Energy Resets':           (68,  ''),
+        'Total Consumption':                     (69,  'kWh', consumption),
+        'Total Consumption Resets':              (70,  ''),
+
+        'AC Power':                              (181, 'W', to_signed),
+
+        'EPS Frequency':                         (63, 'Hz', div100),
+        'EPS Total Energy':                      (110, 'kWh',
+                                                  eps_total_energy),
+        'EPS Total Energy Resets':               (111, 'Hz'),
+    }
+
+    @classmethod
+    def sensor_map(cls):
+        """
+        Return sensor map
+        """
+        sensors = {}
+        for name, (idx, unit, *_) in cls.__sensor_map.items():
+            sensors[name] = (idx, unit)
+        return sensors
+
+    @classmethod
+    def postprocess_map(cls):
+        """
+        Return postprocessing map
+        """
+        sensors = {}
+        for name, (_, _, *processor) in cls.__sensor_map.items():
+            if processor:
+                sensors[name] = processor[0]
+        return sensors
+
+    @classmethod
+    def schema(cls):
+        return cls.__schema
+
+
 class X1(InverterPost):
     __schema = vol.Schema({
         vol.Required('type'): vol.All(
@@ -410,5 +541,71 @@ class X1Mini(InverterPost):
         return cls.__schema
 
 
+class X1MiniV34(InverterPost):
+    __schema = vol.Schema({
+        vol.Required('type', 'type'): vol.All(int, 4),
+        vol.Required('sn',): str,
+        vol.Required('ver'): str,
+        vol.Required('Data'): vol.Schema(
+            vol.All(
+                [vol.Coerce(float)],
+                vol.Any(
+                    vol.Length(min=69, max=69),
+                    vol.Length(min=200, max=200),
+                )
+            )
+        ),
+        vol.Required('Information'): vol.Schema(
+            vol.Any(
+                vol.Length(min=9, max=9),
+                vol.Length(min=10, max=10)
+            )
+        ),
+    }, extra=vol.REMOVE_EXTRA)
+
+    __sensor_map = {
+        'Network Voltage':            (0, 'V', div10),
+        'Output Current':             (1, 'A', div10),
+        'AC Power':                   (2, 'W'),
+        'PV1 Voltage':                (3, 'V', div10),
+        'PV2 Voltage':                (4, 'V', div10),
+        'PV1 Current':                (5, 'A', div10),
+        'PV2 Current':                (6, 'A', div10),
+        'PV1 Power':                  (7, 'W'),
+        'PV2 Power':                  (8, 'W'),
+        'Grid Frequency':             (9, 'Hz', div100),
+        'Total Energy':               (11, 'kWh', div10),
+        'Today\'s Energy':            (13, 'kWh', div10),
+        'Total Feed-in Energy':       (41, 'kWh', div10),
+        'Total Consumption':          (42, 'kWh', div10),
+        'Power Now':                  (43, 'W', div10),
+    }
+
+    @classmethod
+    def sensor_map(cls):
+        """
+        Return sensor map
+        """
+        sensors = {}
+        for name, (idx, unit, *_) in cls.__sensor_map.items():
+            sensors[name] = (idx, unit)
+        return sensors
+
+    @classmethod
+    def postprocess_map(cls):
+        """
+        Return postprocessing map
+        """
+        sensors = {}
+        for name, (_, _, *processor) in cls.__sensor_map.items():
+            if processor:
+                sensors[name] = processor[0]
+        return sensors
+
+    @classmethod
+    def schema(cls):
+        return cls.__schema
+
+
 # registry of inverters
-REGISTRY = [XHybrid, X3, X1, X1Mini]
+REGISTRY = [XHybrid, X3, X3V34, X1, X1Mini, X1MiniV34]
